// (C) Copyright 2004, David M. Blei (blei [at] cs [dot] cmu [dot] edu)

// This file is part of LDA-C.

// LDA-C is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your
// option) any later version.

// LDA-C is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA

#include "lda-alpha.h"
#include <math.h>

/*
 * objective function and its derivatives
 *
 */

double alhood(double a, double ss, int D, int K)
{ return(D * (lgamma(K * a) - K * lgamma(a)) + (a - 1) * ss); }

double d_alhood(double a, double ss, int D, int K)
{ return(D * (K * digamma(K * a) - K * digamma(a)) + ss); }

double d2_alhood(double a, int D, int K)
{ return(D * (K * K * trigamma(K * a) - K * trigamma(a))); }


/*
 * newtons method
 *
 */

double opt_alpha(double *ss, int num_docs, int num_topics, double *alpha)
{
    double *talpha;
    double *g, *q;
    int k, iter = 0, negative_exists = 0;
    double b = 0, z = 0, talpha0 = 0, alpha0 = 0;
    double lambda = 0;
    
    g = (double *) calloc(num_topics, sizeof(double));
    q = (double *) calloc(num_topics, sizeof(double));
    talpha0 = 0;
    
    talpha = malloc(sizeof(double)*num_topics);
    for (k = 0;  k < num_topics; k++) {
	ss[k] /= num_docs;
	talpha[k] = alpha[k];
	talpha0 += talpha[k];
    }

    do {
	double delta = 0;
	double temp1 = 0, temp2 = 0;
	iter++;
	negative_exists = 0;
	for (k = 0;  k < num_topics; k++) {
//	    printf("** %g %g %g\n", talpha0, talpha[k], ss[k]);
	    g[k] = (digamma(talpha0) - digamma(talpha[k]) + ss[k]);
	    q[k] = -1 * trigamma(talpha[k]);
	    temp1 += g[k] / q[k];
	    temp2 += 1 / q[k];
//	    printf("** %g %g\n", g[k], q[k]);
	}
	z = trigamma(talpha0);
	b = temp1 / ((1 / z) + temp2);
	alpha0 = talpha0;
	talpha0 = 0;
	for (k = 0, lambda = 0;  k < num_topics; k++) {
	    alpha[k] = talpha[k];
	    delta = (g[k] - b) / q[k];
	    talpha[k] = talpha[k] - delta; 

	    // Handle for negative alpha values. 
	    if (talpha[k] < 0) {
		printf("Warning: alpha[%d] = %g in iter %d, reverting to %3.3f/%3.5f\n", 
			k, talpha[k], iter, MAX_ALPHA, MIN_ALPHA);
		talpha[k] = 0.01;
		negative_exists = 1;
	    }
	    talpha[k] = fmax(talpha[k], MIN_ALPHA);
	    talpha[k] = fmin(talpha[k], MAX_ALPHA);
	    lambda += -1 * num_docs * delta * g[k];
	    talpha0 += talpha[k];
	}
	if (iter > MAX_ALPHA_ITER && negative_exists == 1) { 
	    printf("WARNING: not able to converge in %d iterations\n", iter);
	    negative_exists = 0;
	}
    }
    while(iter < MAX_ALPHA_ITER && lambda > NEWTON_THRESH);
    
    // Regularize each alpha values. 
    for (k = 0;  k < num_topics; k++) {
	talpha[k] = fmax(talpha[k], MIN_ALPHA);
	talpha[k] = fmin(talpha[k], MAX_ALPHA);
    }

    alpha0 = 0;
    // Store the final value.
    for (k = 0;  k < num_topics; k++) {
	alpha[k] = talpha[k];
	alpha0 += alpha[k];
    }
    free(talpha);
    free(g);
    free(q);

    return alpha0;
}
